from data_gateway_sdk.protocols.water5_v_nero.water5_device_info_ch1 import Water5DeviceInfoCh1Data
from data_gateway_sdk.utils.buf_ref import BufRef


def test_water5_device_info_ch1() -> None:
    case = Water5DeviceInfoCh1Data(value=16.3)  # noqa: E501
    assert case == Water5DeviceInfoCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh1Data(value=0.0)  # noqa: E501
    assert case == Water5DeviceInfoCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh1Data(value=16.3)  # noqa: E501
    assert case == Water5DeviceInfoCh1Data.parse(BufRef(case.serialize()))
    case = Water5DeviceInfoCh1Data(value=4294967.295)  # noqa: E501
    assert case == Water5DeviceInfoCh1Data.parse(BufRef(case.serialize()))



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, tzinfo
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (128 bits)   smpm_ul_device_jupiter_12b_counter_volume
#
# RESULT int:        9234096531137368374233790933
# RESULT bin:  MSB   00000000 00000000 00000000 00000000 00011101 11010110 01000011 11111111 11111111 11111111 11111011 11111111 11111111 11111111 11111001 11010101   LSB
# RESULT hex:  LE    d5 f9 ff ff ff fb ff ff ff 43 d6 1d 00 00 00 00
#
# name                           type    size  value(int)                                                                                                                        data(bits)
# -----------------------------  ------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL           u7         7          85                                                                                                                           1010101
# packet_type_id.0.DFF           bool       1           1                                                                                                                          1
# packet_type_id.1.VAL           u2         2           1                                                                                                                        01
# packet_type_id.1.DFF           bool       1           0                                                                                                                       0
# volume_channel_1               uf32p3    32  2147483647                                                                                       01111111111111111111111111111111
# volume_channel_2               uf32p3    32  2147483647                                                       01111111111111111111111111111111
# battery_volts                  uf8p2      8         200                                               11001000
# temperature                    u7         7          58                                        0111010
# event_reset                    bool       1           1                                       1
# event_low_battery_level        bool       1           1                                      1
# event_low_ambient_temperature  bool       1           1                                     1
# RESERVED                       u35       35           0  00000000000000000000000000000000000


class SmpmUlDeviceJupiter12BCounterVolumeData(Packet):
    volume_channel_1: float
    volume_channel_2: float
    battery_volts: float
    temperature: int
    event_reset: bool
    event_low_battery_level: bool
    event_low_ambient_temperature: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((85) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.volume_channel_1, (int, float))
        result |= ((int(round(float(data.volume_channel_1) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.volume_channel_2, (int, float))
        result |= ((int(round(float(data.volume_channel_2) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 2.55
        result |= ((int(round(float(data.battery_volts) * 100.0, 0))) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_ambient_temperature, bool)
        result |= ((int(data.event_low_ambient_temperature)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceJupiter12BCounterVolumeData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 85 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["volume_channel_1"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp1["volume_channel_2"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp1["battery_volts"] = round(buf.shift(8) / 100.0, 2)
        result__el_tmp1["temperature"] = buf.shift(7) + -35
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp1["event_low_ambient_temperature"] = bool(buf.shift(1))
        result = SmpmUlDeviceJupiter12BCounterVolumeData(**result__el_tmp1)
        buf.shift(35)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)) + Decimal(str(1.8)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)) + Decimal(str(35)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.volume_channel_1)),
                        overloading_value=Decimal(str(17179869183.0)),
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=2,
                        value=Decimal(str(self.volume_channel_2)),
                        overloading_value=Decimal(str(17179869183.0)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.LOW_AMBIENT_TEMPERATURE] if self.event_low_ambient_temperature else []),
                ],
            ),
        ]



from decimal import Decimal
from typing import Dict, List, Any
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, \
    CounterType, ResourceType, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   water5_device_weekly_ch1
#
# RESULT int:        14051230840548075635
# RESULT bin:  MSB   1100001100000000000000000000000010111011111000011010100001110011   LSB
# RESULT hex:  LE    73 a8 e1 bb 00 00 00 c3
#
#
# name           type    size  value(int)                                                        data(bits)
# ---------------------------------------------------------------------------------------------------------
# pack_id        u8         8         115                                                          01110011
# value          uf27p3    27    12313000                               000101110111110000110101000
# UNUSED         -         21           0          000000000000000000000
# battery_fract  u7         7          67   1000011
# battery_int    u1         1           1  1


class Water5DeviceWeeklyCh1Data(Packet):
    value: float
    battery_int: int
    battery_fract: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((115) & (2 ** (8) - 1)) << size
        size += 8
        assert isinstance(data.value, (int, float))
        assert 0.0 <= data.value
        result |= ((int(round(float(data.value) * 1000.0, 0))) & (2 ** (27) - 1)) << size
        size += 27
        result |= ((0) & (2 ** (21) - 1)) << size
        size += 21
        assert isinstance(data.battery_fract, int)
        assert 0 <= data.battery_fract <= 127
        result |= ((data.battery_fract) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_int, int)
        assert 2 <= data.battery_int <= 3
        result |= (((data.battery_int + -2)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'Water5DeviceWeeklyCh1Data':
        result__el_tmp1: Dict[str, Any] = dict()
        if 115 != buf.shift(8):
            raise ValueError("pack_id: buffer doesn't match value")
        result__el_tmp1["value"] = round(buf.shift(27) / 1000.0 - 0.0, 3)
        buf.shift(21)
        result__el_tmp1["battery_fract"] = buf.shift(7) + 0
        result__el_tmp1["battery_int"] = buf.shift(1) + 2
        result = Water5DeviceWeeklyCh1Data(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_int + self.battery_fract / 100)),
                        sensor_type=SensorType.BATTERY,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.value)),
                        overloading_value=Decimal(str(134217.727)),
                    ),
                    # : List[IntegrationV0MessageConsumption]
                ],
            ),
        ]



from data_gateway_sdk.crypto_algorithms.constants.constants_kuz import KUZNECHIK_KEY_LENGTH, KUZNECHIK_BLOCK_SIZE, GOST_PI, GOST_LVEC, GOST_PI_INV, ITER_CONST, MULT_TABLE
from data_gateway_sdk.crypto_algorithms.extensions.errors import KuznechikError
from data_gateway_sdk.crypto_algorithms.extensions.type_alias import Vector, VectorArray


def valid_kuznechik_key(key_bytes: bytes) -> int:
    """
    Валидация ключа шифрования (ключ размером 32 байта)
    """
    if len(key_bytes) == KUZNECHIK_KEY_LENGTH:
        return KUZNECHIK_KEY_LENGTH
    else:
        raise KuznechikError(f"invalid length KEY '{len(key_bytes)}' was given, must be '{KUZNECHIK_KEY_LENGTH}'")


def valid_kuznechik_payload(payload_bytes: bytes) -> int:
    """
    Валидация данных (размер блока 16 байт)
    """
    if len(payload_bytes) == KUZNECHIK_BLOCK_SIZE:
        return KUZNECHIK_BLOCK_SIZE
    else:
        raise KuznechikError(f"invalid length PAYLOAD '{len(payload_bytes)}' was given, must be '{KUZNECHIK_BLOCK_SIZE}'")


def kuz_x(array_a: Vector, array_b: Vector) -> Vector:
    """
    X преобразование. Операция наложения раундового ключа или побитовый XOR ключа и входного блока данных.
    """
    return [array_a[kuz_x_i] ^ array_b[kuz_x_i] for kuz_x_i in range(KUZNECHIK_BLOCK_SIZE)]


def kuz_s(in_data: Vector) -> Vector:
    """
    Нелинейное преобразование S, которое представляет собой простую замену одного байта на другой в соответствии с таблицей.
    """
    out_data: Vector = [0 for _ in range(KUZNECHIK_BLOCK_SIZE)]
    for kuz_s_i in range(KUZNECHIK_BLOCK_SIZE):
        out_data[kuz_s_i] = GOST_PI[in_data[kuz_s_i]]
    return out_data


def sum_field(array: Vector) -> int:
    """
    Вспомогательная функция суммирования всех элементов массива.
    """
    summary = 0
    for element in array:
        summary ^= element
    return summary


def kuz_gf_mul(value_a: int, value_b: int) -> int:
    """
    Вспомогательная функция умножения в поле Галуа
    по полиному  x^8+x^7+x^6+x+1.
    """
    constant = 0
    while value_a:
        if value_a & 1:
            constant ^= value_b
        if value_b & 0x80:
            value_b = (value_b << 1) ^ 0x1C3
        else:
            value_b <<= 1
        value_a >>= 1
    return constant


def l_func(array: Vector) -> int:
    """
    Линейное преобразование. Каждый байт из блока умножается в поле Галуа на один из коэффициентов ряда
    (148, 32, 133, 16, 194, 192, 1, 251, 1, 192, 194, 16, 133, 32, 148, 1)
    в зависимости от порядкового номера байта (ряд представлен для порядковых номеров от 15-ого до 0-ого, как представлено на рисунке).
    Байты складываются между собой по модулю 2, и все 16 байт блока сдвигаются в сторону младшего разряда,
    а полученное число записывается на место считанного байта.
    """
    multiplication = [MULT_TABLE[array[i]][GOST_LVEC[i]] for i in range(len(array))]
    return sum_field(multiplication)


def kuz_r(state: Vector) -> Vector:
    """
    Преобразование R сдвигает данные и реализует уравнение, представленное для расчета L-функции.
    """
    return [l_func(state)] + state[:-1]


def kuz_l(in_data: Vector) -> Vector:
    """
    Линейное преобразование.
    """
    internal = in_data
    for _ in range(KUZNECHIK_BLOCK_SIZE):
        internal = kuz_r(internal)
    return internal


def kuz_reverse_s(in_data: Vector) -> Vector:
    """
    Обратное S преобразовнаие.
    """
    out_data: Vector = [0 for _ in range(len(in_data))]
    for kuz_reverse_s_i in range(KUZNECHIK_BLOCK_SIZE):
        data = in_data[kuz_reverse_s_i]
        if data < 0:
            data = data + 256
        out_data[kuz_reverse_s_i] = GOST_PI_INV[data]
    return out_data


def kuz_reverse_r(state: Vector) -> Vector:
    """
    Обратное R преобразование.
    """
    return state[1:] + [l_func(state[1:] + [state[0]])]


def kuz_reverse_l(in_data: Vector) -> Vector:
    """
    Обратное L преобразование.
    """
    internal = in_data
    for _ in range(KUZNECHIK_BLOCK_SIZE):
        internal = kuz_reverse_r(internal)
    out_data = internal
    return out_data


def kuz_f(in_key: VectorArray, iter_const: Vector) -> VectorArray:
    """
    Функция, выполняющая преобразования ячейки Фейстеля.

    Блок разбивается на две равные части — левую L и правую R.
    Левый подблок L изменяется функцией f с использованием ключа K: X = f(L, K). В качестве функции может быть любое преобразование.
    Полученный подблок X складывается по модулю 2 с правым подблоком R, который остался без изменений: X = X + R.
    Полученные части меняются местами и склеиваются.

    """
    internal = kuz_x(in_key[0], iter_const)
    internal = kuz_s(internal)
    internal = kuz_l(internal)
    out_key_1 = kuz_x(internal, in_key[1])
    return [out_key_1, in_key[0]]


def kuz_expand_key(key: VectorArray) -> VectorArray:
    """
    Функция расчета раундовых ключей

    Итерационные (или раундовые) ключи получаются путем определенных преобразований на основе мастер-ключа,
    длина которого, как мы уже знаем, составляет 256 бит.
    Этот процесс начинается с разбиения мастер-ключа пополам, так получается первая пара раундовых ключей.
    Для генерации каждой последующей пары раундовых ключей применяется восемь итераций сети Фейстеля,
    в каждой итерации используется константа, которая вычисляется путем применения линейного преобразования алгоритма к значению номера итерации.
    """
    round_keys = []
    for expand_i in range(4):
        for k in range(8):
            key = kuz_f(key, ITER_CONST[k + 8 * expand_i])
        round_keys.append(key[0])
        round_keys.append(key[1])
    return round_keys


def _kuz_encrypt(block: Vector, round_keys: VectorArray) -> Vector:
    """
    Функция шифрования блока.
    """
    out_block = block
    for encrypt_i in range(9):
        out_block = kuz_x(round_keys[encrypt_i], out_block)
        out_block = kuz_s(out_block)
        out_block = kuz_l(out_block)
    out_block = kuz_x(out_block, round_keys[9])
    return out_block


def _kuz_decrypt(block: Vector, round_keys: VectorArray) -> Vector:
    """
    Функция расшифровки блока.
    """
    out_block = block
    for decrypt_i in range(9, 0, -1):
        out_block = kuz_x(round_keys[decrypt_i], out_block)
        out_block = kuz_reverse_l(out_block)
        out_block = kuz_reverse_s(out_block)
    out_block = kuz_x(out_block, round_keys[0])
    return out_block


def kuznechik_encryption(plain_text: bytes, key: bytes) -> bytes:
    """
    Обертка функции шифровки блока с разрезом ключа и валидацией входных данных.
    """
    valid_kuznechik_payload(plain_text)
    plain_text_list = list(plain_text)
    valid_kuznechik_key(key)
    key_list = list(key)

    round_key_enc = [key_list[:16], key_list[16:]]
    round_key_enc += kuz_expand_key(round_key_enc)
    result = _kuz_encrypt(plain_text_list, round_key_enc)
    return bytes(result)


def kuznechik_decryption(encrypted_text: bytes, key: bytes) -> bytes:
    """
    Обертка функции расшифровки блока с разрезом ключа и валидацией входных данных.
    """
    valid_kuznechik_payload(encrypted_text)
    encrypted_text_list = list(encrypted_text)
    valid_kuznechik_key(key)
    key_list = list(key)

    round_key_dec = [key_list[:16], key_list[16:]]
    round_key_dec += kuz_expand_key(round_key_dec)
    result = _kuz_decrypt(encrypted_text_list, round_key_dec)
    return bytes(result)



import math
from datetime import datetime, timedelta


START_POINT = datetime(year=2020, month=1, day=1)


def timestamp_calculation(
        current_pack_time: datetime,
        timestamp: timedelta,
        overflow_clock_value_timedelta: timedelta,
        start_point: datetime = START_POINT,
) -> datetime:
    delta = int((current_pack_time.replace(tzinfo=None) - start_point).total_seconds())
    previous_overflow_value = delta // overflow_clock_value_timedelta.total_seconds() * overflow_clock_value_timedelta.total_seconds()
    value_slot_size = math.ceil(math.log2(int(overflow_clock_value_timedelta.total_seconds())))
    value_mask = (2 ** value_slot_size - 1)
    older_mask = (2 ** (64 - value_slot_size) - 1) << value_slot_size
    prev_result_i = int(previous_overflow_value) & (2 ** 64 - 1)
    input_value_i = int(timestamp.total_seconds()) & value_mask
    has_overflow = input_value_i & value_mask < prev_result_i & value_mask
    prev_result_ii = prev_result_i & older_mask
    result_i = prev_result_ii
    result_i |= input_value_i
    if has_overflow:
        result_i += 2 ** value_slot_size - 1
    if result_i > delta + 15 * 60 * 60:  # If device timezone is +14h + max unsynchronized delta(1h)
        result_i -= overflow_clock_value_timedelta.total_seconds()
    elif result_i < delta - 13 * 60 * 60:  # If device timezone is -12h + max unsynchronized delta(1h)
        result_i += overflow_clock_value_timedelta.total_seconds()
    return start_point + timedelta(seconds=result_i)



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, time, tzinfo
from pytz import timezone

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (64 bits)   smpm_ul_device_water_meter_08b_daily
#
# RESULT int:        8663416082350662760
# RESULT bin:  MSB   01111000 00111010 10100011 00000110 10110001 11101010 11100100 01101000   LSB
# RESULT hex:  LE    68 e4 ea b1 06 a3 3a 78
#
# name                                 type    size  value(int)                                                        data(bits)   # noqa: E501
# -----------------------------------  ------  ----  ----------  ----------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL                 u7         7         104                                                           1101000   # noqa: E501
# packet_type_id.0.DFF                 bool       1           0                                                          0          # noqa: E501
# direct_flow_volume                   uf32p3    32   112323300                          00000110101100011110101011100100           # noqa: E501
# direct_flow_volume_day_ago           uf7p1      7          35                   0100011
# event_battery_or_temperature_limits  bool       1           1                  1
# temperature                          u7         7          58           0111010
# event_magnet                         bool       1           0          0
# event_continues_consumption          bool       1           0         0
# event_case_was_opened                bool       1           0        0
# event_system_error                   bool       1           0       0
# event_reset                          bool       1           1      1
# event_sensor_error                   bool       1           1     1
# event_no_resource                    bool       1           1    1
# event_flow_speed_is_over_limit       bool       1           1   1
# event_flow_reverse                   bool       1           0  0


class SmpmUlDeviceWaterMeter08BDailyData(Packet):
    direct_flow_volume: float
    direct_flow_volume_day_ago: float
    event_battery_or_temperature_limits: bool
    temperature: int
    event_magnet: bool
    event_continues_consumption: bool
    event_case_was_opened: bool
    event_system_error: bool
    event_reset: bool
    event_sensor_error: bool
    event_no_resource: bool
    event_flow_speed_is_over_limit: bool
    event_flow_reverse: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((104) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.direct_flow_volume, (int, float))
        result |= ((int(round(float(data.direct_flow_volume) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.direct_flow_volume_day_ago, (int, float))
        result |= ((int(round(max(min(12.7, float(data.direct_flow_volume_day_ago)), 0.0) * 10.0, 0))) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_battery_or_temperature_limits, bool)
        result |= ((int(data.event_battery_or_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_magnet, bool)
        result |= ((int(data.event_magnet)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_continues_consumption, bool)
        result |= ((int(data.event_continues_consumption)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_case_was_opened, bool)
        result |= ((int(data.event_case_was_opened)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_system_error, bool)
        result |= ((int(data.event_system_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error, bool)
        result |= ((int(data.event_sensor_error)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_no_resource, bool)
        result |= ((int(data.event_no_resource)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_speed_is_over_limit, bool)
        result |= ((int(data.event_flow_speed_is_over_limit)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_flow_reverse, bool)
        result |= ((int(data.event_flow_reverse)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceWaterMeter08BDailyData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 104 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["direct_flow_volume"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp1["direct_flow_volume_day_ago"] = round(buf.shift(7) / 10.0, 1)
        result__el_tmp1["event_battery_or_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp1["temperature"] = buf.shift(7) + -35
        result__el_tmp1["event_magnet"] = bool(buf.shift(1))
        result__el_tmp1["event_continues_consumption"] = bool(buf.shift(1))
        result__el_tmp1["event_case_was_opened"] = bool(buf.shift(1))
        result__el_tmp1["event_system_error"] = bool(buf.shift(1))
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_sensor_error"] = bool(buf.shift(1))
        result__el_tmp1["event_no_resource"] = bool(buf.shift(1))
        result__el_tmp1["event_flow_speed_is_over_limit"] = bool(buf.shift(1))
        result__el_tmp1["event_flow_reverse"] = bool(buf.shift(1))
        result = SmpmUlDeviceWaterMeter08BDailyData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        messages = []
        v = int(self.direct_flow_volume_day_ago * 10)
        c = self.direct_flow_volume - self.direct_flow_volume_day_ago
        if v != 0 and v != 127 and c > 0:
            messages.append(IntegrationV0MessageData(
                dt=datetime.combine(received_at.astimezone(device_tz).date(), time(hour=0, minute=0, second=0), device_tz).astimezone(timezone('UTC')),
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(c)),
                        overloading_value=Decimal(str(4294967.295)),
                    ),
                ],
            ))
        return [
            *messages,
            IntegrationV0MessageData(
                dt=received_at,
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.direct_flow_volume)),
                        overloading_value=Decimal(str(4294967.295)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.MAGNET_WAS_DETECTED] if self.event_magnet else []),
                    *([IntegrationV0MessageEvent.BATTERY_OR_TEMPERATURE_LIMITS] if self.event_battery_or_temperature_limits else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.FLOW_REVERSE] if self.event_flow_reverse else []),
                    *([IntegrationV0MessageEvent.ERROR_SYSTEM] if self.event_system_error else []),
                    *([IntegrationV0MessageEvent.CONTINUES_CONSUMPTION] if self.event_continues_consumption else []),
                    *([IntegrationV0MessageEvent.CASE_WAS_OPENED] if self.event_case_was_opened else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR] if self.event_sensor_error else []),
                    *([IntegrationV0MessageEvent.NO_RESOURCE] if self.event_no_resource else []),
                    *([IntegrationV0MessageEvent.FLOW_SPEED_OVER_LIMIT] if self.event_flow_speed_is_over_limit else []),
                ],
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
            ),
        ]



from datetime import datetime, time, tzinfo

from dateutil.relativedelta import relativedelta


def first_day_of_month(dt: datetime, device_tz: tzinfo) -> datetime:
    return datetime.combine(dt + relativedelta(day=1), time(0), device_tz)



from decimal import Decimal
from datetime import timedelta, datetime, time, tzinfo
from typing import List, Any, Dict

from data_aggregator_sdk.constants.enums import JournalDataType, SensorType
from pytz import timezone
from data_aggregator_sdk.integration_message import IntegrationV0MessageEvent, IntegrationV0MessageData, \
    IntegrationV0MessageConsumption, CounterType, ResourceType, IntegrationV0MessageClock, \
    IntegrationV0MessageGeneration, IntegrationV0MessageSensor, IntegrationV0MessageConnectionQuality

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.timestamp_calculation import timestamp_calculation


# PACKET (128 bits)   smpm_ul_device_water_meter_nbiot_16b_daily
#
# RESULT int:        141810285579170350578080716311712763013
# RESULT bin:  MSB   01101010 10101111 10101010 11100100 00000110 10110001 11101010 11100100 10010000 10111010 01111111 11111111 11111111 11000000 00001100 10000101   LSB
# RESULT hex:  LE    85 0c c0 ff ff 7f ba 90 e4 ea b1 06 e4 aa af 6a
#
# name                      type       size  value(int)                                                                                                                        data(bits)
# ------------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------
# packet_type_id.0.VAL      u7            7           5                                                                                                                           0000101
# packet_type_id.0.DFF      bool          1           1                                                                                                                          1
# packet_type_id.1.VAL      u2            2           0                                                                                                                        00
# packet_type_id.1.DFF      bool          1           1                                                                                                                       1
# packet_type_id.2.VAL      u2            2           1                                                                                                                     01
# packet_type_id.2.DFF      bool          1           0                                                                                                                    0
# days_ago                  timedelta     5           0                                                                                                               00000
# sync_time_days_ago        timedelta     3           0                                                                                                            000
# timestamp_s               timedelta    26    33554431                                                                                  01111111111111111111111111
# temperature               u7            7          58                                                                           0111010
# battery_volts             uf6p1         6          33                                                                     100001
# event_reset               bool          1           0                                                                    0
# event_low_battery_level   bool          1           0                                                                   0
# event_temperature_limits  bool          1           1                                                                  1
# direct_flow_volume        uf32p3       32   112323300                                  00000110101100011110101011100100
# reverse_flow_volume       uf12p3       12        2788                      101011100100
# event_magnet              bool          1           0                     0
# rsrp                      u7            7         125              1111101
# rsrq                      u4            4          10          1010
# snr                       u6            6          42    101010
# ecl                       u2            2           1  01


class SmpmUlDeviceWaterMeterNbiot16BDailyData(Packet):
    days_ago: timedelta
    sync_time_days_ago: timedelta
    timestamp_s: timedelta
    temperature: int
    battery_volts: float
    event_reset: bool
    event_low_battery_level: bool
    event_temperature_limits: bool
    direct_flow_volume: float
    reverse_flow_volume: float
    event_magnet: bool
    rsrp: int
    rsrq: int
    snr: int
    ecl: int

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((5) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((0) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((1) & (2 ** (1) - 1)) << size
        size += 1
        result |= ((1) & (2 ** (2) - 1)) << size
        size += 2
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp1 = int(
            data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp1 <= 31
        result |= ((days_ago_tmp1) & (2 ** (5) - 1)) << size
        size += 5
        isinstance(data.sync_time_days_ago, (int, timedelta))
        result |= ((max(min(7,
                            int(data.sync_time_days_ago.total_seconds() // 86400 if isinstance(data.sync_time_days_ago,
                                                                                               timedelta) else data.sync_time_days_ago // 86400)),
                        0)) & (2 ** (3) - 1)) << size
        size += 3
        isinstance(data.timestamp_s, (int, timedelta))
        value_int_tmp2 = int(data.timestamp_s.total_seconds() // 1 if isinstance(data.timestamp_s,
                                                                                 timedelta) else data.timestamp_s // 1) & 67108863
        result |= ((value_int_tmp2) & (2 ** (26) - 1)) << size
        size += 26
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.direct_flow_volume, (int, float))
        result |= ((int(round(float(data.direct_flow_volume) * 1000.0, 0)) & 4294967295) & (2 ** (32) - 1)) << size
        size += 32
        assert isinstance(data.reverse_flow_volume, (int, float))
        result |= ((int(round(float(data.reverse_flow_volume) * 1000.0, 0)) & 4095) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.event_magnet, bool)
        result |= ((int(data.event_magnet)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.rsrp, int)
        assert -140 <= data.rsrp <= -13
        result |= (((data.rsrp + 140)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.rsrq, int)
        assert -19 <= data.rsrq <= -4
        result |= (((data.rsrq + 19)) & (2 ** (4) - 1)) << size
        size += 4
        assert isinstance(data.snr, int)
        assert -20 <= data.snr <= 43
        result |= (((data.snr + 20)) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.ecl, int)
        assert 0 <= data.ecl <= 3
        result |= ((data.ecl) & (2 ** (2) - 1)) << size
        size += 2
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceWaterMeterNbiot16BDailyData':
        result__el_tmp3: Dict[str, Any] = dict()
        if 5 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 1 != buf.shift(2):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp3["days_ago"] = timedelta(seconds=buf.shift(5) * 86400)
        result__el_tmp3["sync_time_days_ago"] = timedelta(seconds=buf.shift(3) * 86400)
        result__el_tmp3["timestamp_s"] = timedelta(seconds=buf.shift(26) * 1)
        result__el_tmp3["temperature"] = buf.shift(7) + -35
        result__el_tmp3["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp3["event_reset"] = bool(buf.shift(1))
        result__el_tmp3["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp3["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp3["direct_flow_volume"] = round(buf.shift(32) / 1000.0, 3)
        result__el_tmp3["reverse_flow_volume"] = round(buf.shift(12) / 1000.0, 3)
        result__el_tmp3["event_magnet"] = bool(buf.shift(1))
        result__el_tmp3["rsrp"] = buf.shift(7) + -140
        result__el_tmp3["rsrq"] = buf.shift(4) + -19
        result__el_tmp3["snr"] = buf.shift(6) + -20
        result__el_tmp3["ecl"] = buf.shift(2) + 0
        result = SmpmUlDeviceWaterMeterNbiot16BDailyData(**result__el_tmp3)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=datetime.combine(
                    (received_at.astimezone(device_tz) - self.days_ago).date(),
                    time(hour=0, minute=0, second=0),
                    device_tz,
                ).astimezone(timezone('UTC')) + timedelta(days=1) if self.days_ago.total_seconds() else received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        channel=1,
                        value=Decimal(str(self.direct_flow_volume)),
                        overloading_value=Decimal(str(4294967.295)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                    *([IntegrationV0MessageEvent.MAGNET_WAS_DETECTED] if self.event_magnet else []),
                ],
                generation=[
                    IntegrationV0MessageGeneration(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.reverse_flow_volume)),
                        overloading_value=Decimal(str(40.95)),
                        journal_data_type=JournalDataType.END_OF_DAY if self.days_ago else JournalDataType.CURRENT,
                    ),
                ],
                clock=[IntegrationV0MessageClock(value=timestamp_calculation(received_at, self.timestamp_s, timedelta(seconds=67108863)))] if self.timestamp_s else [],
                connection_quality=IntegrationV0MessageConnectionQuality(
                    rsrp=self.rsrp,
                    rsrq=self.rsrq,
                    snr=self.snr,
                    ecl=self.ecl,
                )
            ),
        ]



from data_gateway_sdk.protocols.smpm.smpm_ul_device_heat_proxy_meter_16b_daily import SmpmUlDeviceHeatProxyMeter16BDailyData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_heat_proxy_meter_16b_daily() -> None:
    case_serialized = bytes.fromhex("84640000e40c0000030080524d970200")
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=3.3, uptime_min=3, meter_battery_volts=3.3, capacitor_volts=3.33, radio_proxy_battery_volts=3.31, error_meter_sync=False, error_reset=False) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(value=3.3, uptime_min=3, meter_battery_volts=3.3, capacitor_volts=3.33, radio_proxy_battery_volts=3.31, error_meter_sync=False, error_reset=False))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000000000fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00000c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07f00fe0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=0.0, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=0.0, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("846400000000000000000000ffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=0.0, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=0.0, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fff010c00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=0.0, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=0.0, uptime_min=8388608, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=0, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=0, value=268435.456))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=0.0))  # noqa: E501
    case_serialized = bytes.fromhex("84640000000000000000c07fffff0f00")  # overflow
    assert SmpmUlDeviceHeatProxyMeter16BDailyData(value=0.0, uptime_min=0, meter_battery_volts=5.11, capacitor_volts=5.11, radio_proxy_battery_volts=5.11, error_meter_sync=True, error_reset=True) == SmpmUlDeviceHeatProxyMeter16BDailyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceHeatProxyMeter16BDailyData.serialize(SmpmUlDeviceHeatProxyMeter16BDailyData(capacitor_volts=5.11, error_meter_sync=True, error_reset=True, meter_battery_volts=5.11, radio_proxy_battery_volts=5.11, uptime_min=8388608, value=268435.456))  # noqa: E501



from typing import TypeVar, Type, Dict, Any

from data_aggregator_sdk.integration_message import IntegrationV0MessageMeta
from pydantic import BaseModel

T = TypeVar('T', bound='NeroBsPacket')


class NeroBsPacket(BaseModel):

    @classmethod
    def parse(cls: Type[T], data: Dict[str, Any], **kwargs: Any) -> T:
        raise NotImplementedError()

    def to_integration_meta(self, **kwargs: Any) -> IntegrationV0MessageMeta:
        raise NotImplementedError()



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_profile_8h1_energy import \
    SmpmUlDeviceEnergy16BProfile8H1EnergyData, SmpmUlDeviceEnergy16bProfile8h1EnergyType
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_profile_8h1_energy() -> None:
    case_serialized = bytes.fromhex("6a060080002000060001280006e000c8")
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_ACTIVE, point_factor_multiplier=2, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_ACTIVE, point_factor_multiplier=2, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=5.0))  # noqa: E501
    case_serialized = bytes.fromhex("6a030080002000060001280006e00000")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=1, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=0.0), point_factor=0.0, point_factor_multiplier=1, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6a0f0080002000060001280006e00000")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=4, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=0.0), point_factor=0.0, point_factor_multiplier=4, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6a030080002000060001280006e000fc")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=1, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=6.3) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=0.0), point_factor=6.3, point_factor_multiplier=1, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6a0f0080002000060001280006e000fc")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=4, days_ago=timedelta(seconds=0.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=6.3) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=0.0), point_factor=6.3, point_factor_multiplier=4, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6af30380002000060001280006e00000")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=1, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=5443200.0), point_factor=0.0, point_factor_multiplier=1, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6aff0380002000060001280006e00000")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=4, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=0.0) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=5443200.0), point_factor=0.0, point_factor_multiplier=4, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6af30380002000060001280006e000fc")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=1, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=6.3) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=5443200.0), point_factor=6.3, point_factor_multiplier=1, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501
    case_serialized = bytes.fromhex("6aff0380002000060001280006e000fc")  # overflow
    assert SmpmUlDeviceEnergy16BProfile8H1EnergyData(type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE, point_factor_multiplier=4, days_ago=timedelta(seconds=5443200.0), profile=(0, 1, 2, 3, 4, 5, 6, 7), point_factor=6.3) == SmpmUlDeviceEnergy16BProfile8H1EnergyData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BProfile8H1EnergyData.serialize(SmpmUlDeviceEnergy16BProfile8H1EnergyData(days_ago=timedelta(seconds=5443200.0), point_factor=6.3, point_factor_multiplier=4, profile=(0, 1, 2, 3, 4, 5, 6, 7), type=SmpmUlDeviceEnergy16bProfile8h1EnergyType.ENERGY_CONSUMED_REACTIVE))  # noqa: E501



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, tzinfo

from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent, SensorType
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageSensor

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet


# PACKET (32 bits)   smpm_ul_device_water_meter_04b_event
#
# RESULT int:        104521996
# RESULT bin:  MSB   00000110 00111010 11100001 00001100   LSB
# RESULT hex:  LE    0c e1 3a 06
#
# name                      type   size  value(int)                        data(bits)
# ------------------------  -----  ----  ----------  --------------------------------
# packet_type_id.0.VAL      u7        7          12                           0001100
# packet_type_id.0.DFF      bool      1           0                          0
# battery_volts             uf6p1     6          33                    100001
# event_low_battery_level   bool      1           1                   1
# event_temperature_limits  bool      1           1                  1
# temperature               u7        7          58           0111010
# event_case_was_opened     bool      1           0          0
# event_magnet              bool      1           0         0
# event_reset               bool      1           1        1
# event_sensor_error        bool      1           1       1
# RESERVED                  u5        5           0  00000


class SmpmUlDeviceWaterMeter04BEventData(Packet):
    battery_volts: float
    event_low_battery_level: bool
    event_temperature_limits: bool
    temperature: int
    event_case_was_opened: bool
    event_magnet: bool
    event_reset: bool
    event_sensor_error: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((12) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.battery_volts, (int, float))
        assert 0.0 <= data.battery_volts <= 6.3
        result |= ((int(round(float(data.battery_volts) * 10.0, 0))) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_temperature_limits, bool)
        result |= ((int(data.event_temperature_limits)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.temperature, int)
        assert -35 <= data.temperature <= 92
        result |= (((data.temperature + 35)) & (2 ** (7) - 1)) << size
        size += 7
        assert isinstance(data.event_case_was_opened, bool)
        result |= ((int(data.event_case_was_opened)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_magnet, bool)
        result |= ((int(data.event_magnet)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_sensor_error, bool)
        result |= ((int(data.event_sensor_error)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(4, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceWaterMeter04BEventData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 12 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["battery_volts"] = round(buf.shift(6) / 10.0, 1)
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result__el_tmp1["event_temperature_limits"] = bool(buf.shift(1))
        result__el_tmp1["temperature"] = buf.shift(7) + -35
        result__el_tmp1["event_case_was_opened"] = bool(buf.shift(1))
        result__el_tmp1["event_magnet"] = bool(buf.shift(1))
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_sensor_error"] = bool(buf.shift(1))
        result = SmpmUlDeviceWaterMeter04BEventData(**result__el_tmp1)
        buf.shift(5)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                sensors=[
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.battery_volts)),
                        sensor_type=SensorType.BATTERY,
                    ),
                    IntegrationV0MessageSensor(
                        value=Decimal(str(self.temperature)),
                        sensor_type=SensorType.TEMPERATURE,
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.MAGNET_WAS_DETECTED] if self.event_magnet else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.CASE_WAS_OPENED] if self.event_case_was_opened else []),
                    *([IntegrationV0MessageEvent.ERROR_SENSOR] if self.event_sensor_error else []),
                    *([IntegrationV0MessageEvent.TEMPERATURE_LIMIT] if self.event_temperature_limits else []),
                ],
            ),
        ]



from datetime import timedelta, datetime, tzinfo, time
from enum import IntEnum, unique
from typing import Tuple, Any, Dict, List, Optional

from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageProfile, ProfileGranulation, GRANULATION_TO_END_OF_DATETIME_MAP, ProfileKind
from data_aggregator_sdk.utils.round_dt import round_dt

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.days_ago_calculation import days_ago_calculation
from data_gateway_sdk.utils.packet import Packet
from data_gateway_sdk.utils.true_round import true_round


# PACKET (128 bits)   smpm_ul_device_energy_16b_profile_8h_ar
#
# RESULT int:        4153979493411694417272393125779737729
# RESULT bin:  MSB   00000011 00100000 00000111 00000000 01100000 00000101 00000000 01000000 00000011 00000000 00100000 00000001 00000000 00000000 00001100 10000001   LSB
# RESULT hex:  LE    81 0c 00 00 01 20 00 03 40 00 05 60 00 07 20 03
#
# name                  type       size  value(int)                                                                                                                        data(bits)   # noqa: E501
# --------------------  ---------  ----  ----------  --------------------------------------------------------------------------------------------------------------------------------   # noqa: E501
# packet_type_id.0.VAL  u7            7           1                                                                                                                           0000001   # noqa: E501
# packet_type_id.0.DFF  bool          1           1                                                                                                                          1
# packet_type_id.1.VAL  u2            2           0                                                                                                                        00
# packet_type_id.1.DFF  bool          1           1                                                                                                                       1
# packet_type_id.2.VAL  u2            2           1                                                                                                                     01
# packet_type_id.2.DFF  bool          1           0                                                                                                                    0
# days_ago              timedelta     6           0                                                                                                              000000
# profile.0.point       u12          12           0                                                                                                  000000000000
# profile.1.point       u12          12           1                                                                                      000000000001
# profile.2.point       u12          12           2                                                                          000000000010
# profile.3.point       u12          12           3                                                              000000000011
# profile.4.point       u12          12           4                                                  000000000100
# profile.5.point       u12          12           5                                      000000000101
# profile.6.point       u12          12           6                          000000000110
# profile.7.point       u12          12           7              000000000111
# point_factor          uf12p1       12          50  000000110010
@unique
class SmpmUlDeviceEnergy16bProfile8hARIds(IntEnum):
    UL_DATA_16B__PROFILE_H8_1_AP = 513
    UL_DATA_16B__PROFILE_H8_1_AP__PHASE_A = 514
    UL_DATA_16B__PROFILE_H8_1_AP__PHASE_B = 515
    UL_DATA_16B__PROFILE_H8_1_AP__PHASE_C = 516
    UL_DATA_16B__PROFILE_H8_1_AN = 517
    UL_DATA_16B__PROFILE_H8_1_AN__PHASE_A = 518
    UL_DATA_16B__PROFILE_H8_1_AN__PHASE_B = 519
    UL_DATA_16B__PROFILE_H8_1_AN__PHASE_C = 520
    UL_DATA_16B__PROFILE_H8_1_RP = 521
    UL_DATA_16B__PROFILE_H8_1_RP__PHASE_A = 522
    UL_DATA_16B__PROFILE_H8_1_RP__PHASE_B = 523
    UL_DATA_16B__PROFILE_H8_1_RP__PHASE_C = 524
    UL_DATA_16B__PROFILE_H8_1_RN = 525
    UL_DATA_16B__PROFILE_H8_1_RN__PHASE_A = 526
    UL_DATA_16B__PROFILE_H8_1_RN__PHASE_B = 527
    UL_DATA_16B__PROFILE_H8_1_RN__PHASE_C = 528
    UL_DATA_16B__PROFILE_H8_2_AP = 529
    UL_DATA_16B__PROFILE_H8_2_AP__PHASE_A = 530
    UL_DATA_16B__PROFILE_H8_2_AP__PHASE_B = 531
    UL_DATA_16B__PROFILE_H8_2_AP__PHASE_C = 532
    UL_DATA_16B__PROFILE_H8_2_AN = 533
    UL_DATA_16B__PROFILE_H8_2_AN__PHASE_A = 534
    UL_DATA_16B__PROFILE_H8_2_AN__PHASE_B = 535
    UL_DATA_16B__PROFILE_H8_2_AN__PHASE_C = 536
    UL_DATA_16B__PROFILE_H8_2_RP = 537
    UL_DATA_16B__PROFILE_H8_2_RP__PHASE_A = 538
    UL_DATA_16B__PROFILE_H8_2_RP__PHASE_B = 539
    UL_DATA_16B__PROFILE_H8_2_RP__PHASE_C = 540
    UL_DATA_16B__PROFILE_H8_2_RN = 541
    UL_DATA_16B__PROFILE_H8_2_RN__PHASE_A = 542
    UL_DATA_16B__PROFILE_H8_2_RN__PHASE_B = 543
    UL_DATA_16B__PROFILE_H8_2_RN__PHASE_C = 544
    UL_DATA_16B__PROFILE_H8_3_AP = 545
    UL_DATA_16B__PROFILE_H8_3_AP__PHASE_A = 546
    UL_DATA_16B__PROFILE_H8_3_AP__PHASE_B = 547
    UL_DATA_16B__PROFILE_H8_3_AP__PHASE_C = 548
    UL_DATA_16B__PROFILE_H8_3_AN = 549
    UL_DATA_16B__PROFILE_H8_3_AN__PHASE_A = 550
    UL_DATA_16B__PROFILE_H8_3_AN__PHASE_B = 551
    UL_DATA_16B__PROFILE_H8_3_AN__PHASE_C = 552
    UL_DATA_16B__PROFILE_H8_3_RP = 553
    UL_DATA_16B__PROFILE_H8_3_RP__PHASE_A = 554
    UL_DATA_16B__PROFILE_H8_3_RP__PHASE_B = 555
    UL_DATA_16B__PROFILE_H8_3_RP__PHASE_C = 556
    UL_DATA_16B__PROFILE_H8_3_RN = 557
    UL_DATA_16B__PROFILE_H8_3_RN__PHASE_A = 558
    UL_DATA_16B__PROFILE_H8_3_RN__PHASE_B = 559
    UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C = 560


PROFILE_H8_AR_TYPE_MAP = {
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP__PHASE_A: ProfileKind.ENERGY_A_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP__PHASE_B: ProfileKind.ENERGY_A_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AP__PHASE_C: ProfileKind.ENERGY_A_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AN: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AN__PHASE_A: ProfileKind.ENERGY_A_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AN__PHASE_B: ProfileKind.ENERGY_A_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_AN__PHASE_C: ProfileKind.ENERGY_A_N_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RP: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RP__PHASE_A: ProfileKind.ENERGY_R_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RP__PHASE_B: ProfileKind.ENERGY_R_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RP__PHASE_C: ProfileKind.ENERGY_R_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RN: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RN__PHASE_A: ProfileKind.ENERGY_R_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RN__PHASE_B: ProfileKind.ENERGY_R_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_1_RN__PHASE_C: ProfileKind.ENERGY_R_N_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AP: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AP__PHASE_A: ProfileKind.ENERGY_A_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AP__PHASE_B: ProfileKind.ENERGY_A_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AP__PHASE_C: ProfileKind.ENERGY_A_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AN: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AN__PHASE_A: ProfileKind.ENERGY_A_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AN__PHASE_B: ProfileKind.ENERGY_A_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_AN__PHASE_C: ProfileKind.ENERGY_A_N_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RP: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RP__PHASE_A: ProfileKind.ENERGY_R_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RP__PHASE_B: ProfileKind.ENERGY_R_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RP__PHASE_C: ProfileKind.ENERGY_R_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RN: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RN__PHASE_A: ProfileKind.ENERGY_R_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RN__PHASE_B: ProfileKind.ENERGY_R_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_2_RN__PHASE_C: ProfileKind.ENERGY_R_N_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AP: ProfileKind.ENERGY_A_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AP__PHASE_A: ProfileKind.ENERGY_A_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AP__PHASE_B: ProfileKind.ENERGY_A_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AP__PHASE_C: ProfileKind.ENERGY_A_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AN: ProfileKind.ENERGY_A_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AN__PHASE_A: ProfileKind.ENERGY_A_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AN__PHASE_B: ProfileKind.ENERGY_A_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_AN__PHASE_C: ProfileKind.ENERGY_A_N_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RP: ProfileKind.ENERGY_R_P_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RP__PHASE_A: ProfileKind.ENERGY_R_P_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RP__PHASE_B: ProfileKind.ENERGY_R_P_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RP__PHASE_C: ProfileKind.ENERGY_R_P_C_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN: ProfileKind.ENERGY_R_N_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_A: ProfileKind.ENERGY_R_N_A_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_B: ProfileKind.ENERGY_R_N_B_DELTA,
    SmpmUlDeviceEnergy16bProfile8hARIds.UL_DATA_16B__PROFILE_H8_3_RN__PHASE_C: ProfileKind.ENERGY_R_N_C_DELTA,
}


class SmpmUlDeviceEnergy16BProfile8HArData(Packet):
    packet_type_id: SmpmUlDeviceEnergy16bProfile8hARIds
    days_ago: timedelta
    profile: Tuple[int, int, int, int, int, int, int, int]
    point_factor: float

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        packet_type_id__value_int_tmp1 = 0
        packet_type_id__value_size_tmp2 = 0
        assert isinstance(data.packet_type_id, SmpmUlDeviceEnergy16bProfile8hARIds)
        packet_type_id__value_int_tmp1 |= ((data.packet_type_id.value) & (2 ** (10) - 1)) << packet_type_id__value_size_tmp2
        packet_type_id__value_size_tmp2 += 10
        packet_type_id__max_size_tmp3 = 0
        packet_type_id__steps_tmp4 = [8, 3, 3, 3]
        for packet_type_id__j_tmp5 in range(32):
            packet_type_id__step_tmp6 = packet_type_id__steps_tmp4[packet_type_id__j_tmp5] if packet_type_id__j_tmp5 < len(packet_type_id__steps_tmp4) else packet_type_id__steps_tmp4[-1]  # noqa: E501
            packet_type_id__max_size_tmp3 += packet_type_id__step_tmp6
            packet_type_id__current_part_value_tmp7 = packet_type_id__value_int_tmp1 & (2 ** packet_type_id__step_tmp6 - 1)
            packet_type_id__value_int_tmp1 = packet_type_id__value_int_tmp1 >> (packet_type_id__step_tmp6 - 1)
            result |= ((packet_type_id__current_part_value_tmp7) & (2 ** ((packet_type_id__step_tmp6 - 1)) - 1)) << size
            size += (packet_type_id__step_tmp6 - 1)
            assert isinstance((packet_type_id__value_int_tmp1 != 0), bool)
            result |= ((int((packet_type_id__value_int_tmp1 != 0))) & (2 ** (1) - 1)) << size
            size += 1
            if packet_type_id__value_int_tmp1 == 0:
                break
        isinstance(data.days_ago, (int, timedelta))
        days_ago_tmp8 = int(data.days_ago.total_seconds() // 86400 if isinstance(data.days_ago, timedelta) else data.days_ago // 86400)
        assert 0 <= days_ago_tmp8 <= 63
        result |= ((days_ago_tmp8) & (2 ** (6) - 1)) << size
        size += 6
        assert isinstance(data.profile, tuple) and len(data.profile) == 8
        assert isinstance(data.profile[0], int)
        assert 0 <= data.profile[0] <= 4095
        result |= ((data.profile[0]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[1], int)
        assert 0 <= data.profile[1] <= 4095
        result |= ((data.profile[1]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[2], int)
        assert 0 <= data.profile[2] <= 4095
        result |= ((data.profile[2]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[3], int)
        assert 0 <= data.profile[3] <= 4095
        result |= ((data.profile[3]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[4], int)
        assert 0 <= data.profile[4] <= 4095
        result |= ((data.profile[4]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[5], int)
        assert 0 <= data.profile[5] <= 4095
        result |= ((data.profile[5]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[6], int)
        assert 0 <= data.profile[6] <= 4095
        result |= ((data.profile[6]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.profile[7], int)
        assert 0 <= data.profile[7] <= 4095
        result |= ((data.profile[7]) & (2 ** (12) - 1)) << size
        size += 12
        assert isinstance(data.point_factor, (int, float))
        assert 0.0 <= data.point_factor <= 409.5
        result |= ((int(round(float(data.point_factor) * 10.0, 0))) & (2 ** (12) - 1)) << size
        size += 12
        return result.to_bytes(16, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceEnergy16BProfile8HArData':
        result__el_tmp9: Dict[str, Any] = dict()
        packet_type_id__res_tmp10 = 0
        packet_type_id__steps_tmp11 = (8, 3, 3, 3)
        packet_type_id__res_size_tmp12 = 0
        packet_type_id__step_tmp13 = 0
        for packet_type_id__i_tmp15 in range(32):
            packet_type_id__step_tmp13 = (packet_type_id__steps_tmp11[packet_type_id__i_tmp15] if packet_type_id__i_tmp15 < len(packet_type_id__steps_tmp11) else packet_type_id__steps_tmp11[-1]) - 1  # noqa: E501
            packet_type_id__res_tmp10 |= buf.shift(packet_type_id__step_tmp13) << packet_type_id__res_size_tmp12
            packet_type_id__res_size_tmp12 += packet_type_id__step_tmp13
            packet_type_id__dff_tmp14 = bool(buf.shift(1))
            if not packet_type_id__dff_tmp14:
                break
        packet_type_id__buf_tmp16 = buf
        buf = BufRef(packet_type_id__res_tmp10, stop_on_buffer_end=True)
        result__el_tmp9["packet_type_id"] = SmpmUlDeviceEnergy16bProfile8hARIds(buf.shift(10))
        buf = packet_type_id__buf_tmp16
        result__el_tmp9["days_ago"] = timedelta(seconds=buf.shift(6) * 86400)
        profile_tmp17: List[int] = []
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        profile__item_tmp18 = buf.shift(12) + 0
        profile_tmp17.append(profile__item_tmp18)
        result__el_tmp9["profile"] = tuple(profile_tmp17)
        result__el_tmp9["point_factor"] = round(buf.shift(12) / 10.0, 1)
        result = SmpmUlDeviceEnergy16BProfile8HArData(**result__el_tmp9)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        values: Tuple[Optional[float], ...]
        precision = 2
        profiles = (
            true_round(self.profile[0] * self.point_factor, precision) / 1000,
            true_round(self.profile[1] * self.point_factor, precision) / 1000,
            true_round(self.profile[2] * self.point_factor, precision) / 1000,
            true_round(self.profile[3] * self.point_factor, precision) / 1000,
            true_round(self.profile[4] * self.point_factor, precision) / 1000,
            true_round(self.profile[5] * self.point_factor, precision) / 1000,
            true_round(self.profile[6] * self.point_factor, precision) / 1000,
            true_round(self.profile[7] * self.point_factor, precision) / 1000,
        )
        if self.packet_type_id in {513, 514, 515, 516, 517, 518, 519, 520, 521, 522, 523, 524, 525, 526, 527, 528}:
            values = (
                *profiles,
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
            )
        elif self.packet_type_id in {529, 530, 531, 532, 533, 534, 535, 536, 537, 538, 539, 540, 541, 542, 543, 544}:
            values = (
                None, None, None, None, None, None, None, None,
                *profiles,
                None, None, None, None, None, None, None, None,
            )
        else:
            values = (
                None, None, None, None, None, None, None, None,
                None, None, None, None, None, None, None, None,
                *profiles,
            )
        return [
            IntegrationV0MessageData(
                dt=round_dt(
                    days_ago_calculation(
                        received_at,
                        device_tz,
                        time(0),
                        self.days_ago,
                    ),
                    GRANULATION_TO_END_OF_DATETIME_MAP[ProfileGranulation.MINUTE_60],
                ),
                profiles=[
                    IntegrationV0MessageProfile(
                        type=PROFILE_H8_AR_TYPE_MAP[self.packet_type_id],
                        granulation=ProfileGranulation.MINUTE_60,
                        values=values,
                    ),
                ],
            ),
        ]



from datetime import timedelta

from data_gateway_sdk.protocols.smpm.smpm_ul_device_energy_16b_tariff_generated import \
    SmpmUlDeviceEnergy16BTariffGeneratedData
from data_gateway_sdk.utils.buf_ref import BufRef


def test_smpm_ul_device_energy_16b_tariff_generated() -> None:
    case_serialized = bytes.fromhex("c10250320f1e60873948490500000000")
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=False, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=123123, slot_1=4312123, slot_2=5413, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=False, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=123123, slot_1=4312123, slot_2=5413, slot_3=0))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1025802000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=0.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=0.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=0, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=0, slot_1=67108864, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=0, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=0, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=0, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=67108864, slot_3=0, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501
    case_serialized = bytes.fromhex("c1f25f02000000000000000000000000")  # overflow
    assert SmpmUlDeviceEnergy16BTariffGeneratedData(energy_is_reactive=False, days_ago=timedelta(seconds=10972800.0), valid=True, tariff_mask=(True, False, True, False, False, True, False, False), slot_0=0, slot_1=0, slot_2=0, slot_3=0) == SmpmUlDeviceEnergy16BTariffGeneratedData.parse(BufRef(case_serialized))  # noqa: E501
    assert case_serialized == SmpmUlDeviceEnergy16BTariffGeneratedData.serialize(SmpmUlDeviceEnergy16BTariffGeneratedData(days_ago=timedelta(seconds=10972800.0), energy_is_reactive=False, slot_0=67108864, slot_1=67108864, slot_2=67108864, slot_3=67108864, tariff_mask=(True, False, True, False, False, True, False, False), valid=True))  # noqa: E501



from decimal import Decimal
from typing import List, Any, Dict
from datetime import datetime, tzinfo
from data_aggregator_sdk.constants.enums import IntegrationV0MessageEvent
from data_aggregator_sdk.integration_message import IntegrationV0MessageData, IntegrationV0MessageConsumption, \
    CounterType, ResourceType

from data_gateway_sdk.utils.buf_ref import BufRef
from data_gateway_sdk.utils.packet import Packet

# PACKET (64 bits)   smpm_ul_device_jupiter_08b_counter
#
# RESULT int:        16140901047315988325
# RESULT bin:  MSB   11011111 11111111 11111111 11111011 11111111 11111111 11111111 01100101   LSB
# RESULT hex:  LE    65 ff ff ff fb ff ff df
#
# name                     type  size  value(int)                                                        data(bits)
# -----------------------  ----  ----  ----------  ----------------------------------------------------------------
# packet_type_id.0.VAL     u7       7         101                                                           1100101
# packet_type_id.0.DFF     bool     1           0                                                          0
# value_channel_1          u27     27    67108863                               011111111111111111111111111
# value_channel_2          u27     27    67108863    011111111111111111111111111
# event_reset              bool     1           1   1
# event_low_battery_level  bool     1           1  1


class SmpmUlDeviceJupiter08BCounterData(Packet):
    value_channel_1: int
    value_channel_2: int
    event_reset: bool
    event_low_battery_level: bool

    def serialize(self) -> bytes:
        data = self
        result = 0
        size = 0
        result |= ((101) & (2 ** (7) - 1)) << size
        size += 7
        result |= ((0) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.value_channel_1, int)
        result |= (((data.value_channel_1) & 134217727) & (2 ** (27) - 1)) << size
        size += 27
        assert isinstance(data.value_channel_2, int)
        result |= (((data.value_channel_2) & 134217727) & (2 ** (27) - 1)) << size
        size += 27
        assert isinstance(data.event_reset, bool)
        result |= ((int(data.event_reset)) & (2 ** (1) - 1)) << size
        size += 1
        assert isinstance(data.event_low_battery_level, bool)
        result |= ((int(data.event_low_battery_level)) & (2 ** (1) - 1)) << size
        size += 1
        return result.to_bytes(8, "little")

    @classmethod
    def parse(cls, buf: BufRef) -> 'SmpmUlDeviceJupiter08BCounterData':
        result__el_tmp1: Dict[str, Any] = dict()
        if 101 != buf.shift(7):
            raise ValueError("packet_type_id: buffer doesn't match value")
        if 0 != buf.shift(1):
            raise ValueError("packet_type_id: buffer doesn't match value")
        result__el_tmp1["value_channel_1"] = buf.shift(27) + 0
        result__el_tmp1["value_channel_2"] = buf.shift(27) + 0
        result__el_tmp1["event_reset"] = bool(buf.shift(1))
        result__el_tmp1["event_low_battery_level"] = bool(buf.shift(1))
        result = SmpmUlDeviceJupiter08BCounterData(**result__el_tmp1)
        return result

    def to_integration_data(self, received_at: datetime, device_tz: tzinfo, **kwargs: Any) -> List[IntegrationV0MessageData]:
        return [
            IntegrationV0MessageData(
                dt=received_at,
                consumption=[
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.value_channel_1)),
                        channel=1,
                        overloading_value=Decimal(str(268435455.0)),
                    ),
                    IntegrationV0MessageConsumption(
                        counter_type=CounterType.COMMON,
                        resource_type=ResourceType.COMMON,
                        value=Decimal(str(self.value_channel_2)),
                        channel=2,
                        overloading_value=Decimal(str(268435455.0)),
                    ),
                ],
                events=[
                    *([IntegrationV0MessageEvent.RESET] if self.event_reset else []),
                    *([IntegrationV0MessageEvent.BATTERY_IS_LOW] if self.event_low_battery_level else []),
                ],
            ),
        ]
